"""
Core functionality for wallet finding operations.
"""

import os
import csv
import uuid
import json
import platform
import itertools
import multiprocessing
from pathlib import Path
from multiprocessing.pool import AsyncResult

import requests
from more_itertools import chunked
from .crypto_wallet import CryptoWallet



class Core:
    def __init__(self):
        self.valid_addresses = 0
        self.config = {}
        self.app_dir_path = Path.home() / ".wallet_finder"
        self.app_dir_path.mkdir(parents=True, exist_ok=True)
        self.app_download_path = Path.home() / "Downloads"
        self.config_file = self.app_dir_path / "config.json"

        with open('wordlist.txt', 'r', encoding='utf-8') as file:
            self.wordlist = file.read().splitlines()

        with open('target_addresses.txt', 'r', encoding='utf-8') as file:
            self.target_addresses = file.read().splitlines()

        with open(self.app_download_path / 'output.csv', 'w', encoding='utf-8') as file:
            writer = csv.writer(file)
            writer.writerow(["Seed Phrase", "TRX Address"])

    def check_combination(self, seeds):
        """
        Check if the combination of seeds generates a valid TRX address
        """
        seed_phrase = ' '.join(seeds)
        try:
            wallet = CryptoWallet(seed_phrase)
            for index in range(5):
                address = wallet.get_trx_address(index)
                if address in self.target_addresses:
                    return (seed_phrase, address)
        except ValueError:
            pass
        return None

    def start_process(self) -> None:
        """
        Start the process of checking all combinations
        """

        if self.config["progress"] > 0:
            resume = input("Do you want to continue from the last progress? Press Enter Y/N (N): ")

            if resume.lower() != "y":
                self.config["progress"] = 0
                self.save_config()

        if self.config["progress"] < 0:
            self.config["progress"] = 0
            self.save_config()

        num_cores = multiprocessing.cpu_count()
        chunk_size = num_cores * 5000
        start_from = self.config["progress"]

        combinations = itertools.permutations(self.wordlist, 12)
        with multiprocessing.Pool(processes=num_cores) as pool:
            csv_file = open(self.app_download_path / 'output.csv', 'a', encoding='utf-8')
            for idx, chunk in enumerate(chunked(itertools.islice(combinations, start_from, None), chunk_size), start=int(start_from / chunk_size * 5) - 1):
                process_count = (idx + 1) * chunk_size * 5
                print(f"Checking: {process_count}")
                results: AsyncResult = pool.starmap_async(self.check_combination, [(seeds,) for seeds in chunk])

                results.wait()
                process_results = results.get()
                for result in process_results:
                    if not result:
                        continue
                    
                    seed_phrase, address = result
                    print(f"Found address: {address} with seed: {seed_phrase}")
                    csv_file_writer = csv.writer(csv_file)
                    csv_file_writer.writerow([seed_phrase, address])
                    self.valid_addresses += 1

                if self.valid_addresses >= len(self.target_addresses):
                    pool.terminate()
                    break

                self.config["progress"] = process_count
                self.save_config()
            
            csv_file.close()

    def validate_device(self) -> bool:
        """
        Validate the device with the API key.

        This function sends a POST request to the API to validate the device with the provided API key.


        Returns:
            bool: True if the device is validated; False otherwise.

        Raises:
            Exception: If the request fails or an error occurs during the validation process
        """
        request_data = {
            "device_id": self.config.get("device_id"),
            "device_mac": self.config.get("device_mac"),
            "device_name": self.config.get("device_name")
        }

        if self.config.get("api_key"):
            request_data["api_key"] = self.config.get("api_key")

        try:
            response = requests.post("https://us-central1-crypto-wallet-recovery.cloudfunctions.net/gcp-wallet-finder-validate-device", json=request_data, timeout=120)
        except requests.exceptions.RequestException as e:
            print("Device validation failed: %s", e)
            raise ConnectionError("Device validation failed. Please check your internet connection.")
        
        response_data = response.json()
        print(f"Device validation response: {json.dumps(response_data, indent=4)}")

        if response.status_code == 201:
            self.config["api_key"] = response_data.get("api_key")
            self.save_config()

        return response_data.get("success", False), response_data.get("message", "Unknown Error"), response.status_code

    def get_config(self) -> None:
        """
        Retrieve configuration data from the config.json file.

        If the configuration file does not exist, it creates a new one with default settings.

        Returns:
            dict: The loaded or default configuration data.
        """
        if os.path.exists(self.config_file):
            with open(self.config_file, "r", encoding="utf-8") as f:
                self.config = json.load(f)
        else:
            # Save default configuration to the file
            self.config = {
                "device_id": str(uuid.uuid4()),
                "device_mac": hex(uuid.getnode()),
                "device_name": platform.node(),
                "progress": 0,
                "wordlist_file": ""
            }
            self.save_config()

    def save_config(self) -> None:
        """
        Save the current configuration data to the config.json file.

        This function updates the configuration file with the current settings.
        """
        with open(self.config_file, "w", encoding="utf-8") as f:
            json.dump(self.config, f, indent=4)

    def main(self) -> None:
        """
        Main function to start the wallet finder process.
        """
        multiprocessing.freeze_support()
        self.get_config()
        print(f"Loaded configuration: {json.dumps(self.config, indent=4)}")
        status, msg, _status_code = self.validate_device()
        if not status:
            print(f"Device validation failed: {msg}")
            exit(1)

        self.start_process()
